@@ -11,7 +11,7 @@ use anyhow::Context;
 use console::Term;
 use dialoguer::{theme, Confirm, Input};
 use itertools::Itertools;
-use serde::{de::DeserializeOwned, Serialize};
+use serde::{de, Deserialize, Serialize};
 use structopt::StructOpt;
 // use structopt::clap;
 
@@ -25,9 +25,11 @@ use crate::Output;
 
 use config::Config;
 use helper::AddLocation;
+use print::{ClusterAndStates, Print, StateAndClusters};
 
 mod config;
 mod helper;
+mod print;
 
 const ABOUT: &str = "statehub CLI tool";
 
@@ -395,7 +397,7 @@ impl Cli {
                 let default_state = if no_default_storage_class {
                     None
                 } else if default_storage_class.is_none() {
-                    states.first().map(|state| state.to_string())
+                    states.first().map(ToString::to_string)
                 } else {
                     default_storage_class
                 };
@@ -548,22 +550,22 @@ impl StateHub {
         self.api
             .create_state(state)
             .await
-            .handle_output(&self.stdout, self.json)
+            .print(&self.stdout, self.json)
     }
 
     async fn delete_state(&self, name: v0::StateName) -> anyhow::Result<()> {
         self.api
             .delete_state(name)
             .await
-            .handle_output(&self.stdout, self.json)
+            .print(&self.stdout, self.json)
     }
 
     async fn show_state(self, state: &v0::StateName) -> anyhow::Result<()> {
         let state = self.api.get_state(state).await.map(Detailed)?;
         if let Ok(clusters) = self.api.get_all_clusters().await {
-            (state, clusters).handle_output(&self.stdout, self.json)
+            StateAndClusters::new(state, clusters).print(&self.stdout, self.json)
         } else {
-            state.handle_output(&self.stdout, self.json)
+            state.print(&self.stdout, self.json)
         }
     }
 
@@ -571,14 +573,14 @@ impl StateHub {
         self.api
             .get_all_states()
             .await
-            .handle_output(&self.stdout, self.json)
+            .print(&self.stdout, self.json)
     }
 
     async fn list_clusters(&self) -> anyhow::Result<()> {
         self.api
             .get_all_clusters()
             .await
-            .handle_output(&self.stdout, self.json)
+            .print(&self.stdout, self.json)
     }
 
     async fn register_cluster(
@@ -634,15 +636,15 @@ impl StateHub {
             self.claim_unowned_states_helper(&cluster, states).await?;
         }
 
-        cluster.handle_output(&self.stdout, self.json)
+        cluster.print(&self.stdout, self.json)
     }
 
     async fn show_cluster(&self, name: v0::ClusterName) -> anyhow::Result<()> {
         let cluster = self.api.get_cluster(&name).await.map(Detailed)?;
         if let Ok(states) = self.api.get_all_states().await {
-            (cluster, states).handle_output(&self.stdout, self.json)
+            ClusterAndStates::new(cluster, states).print(&self.stdout, self.json)
         } else {
-            cluster.handle_output(&self.stdout, self.json)
+            cluster.print(&self.stdout, self.json)
         }
     }
 
@@ -656,7 +658,7 @@ impl StateHub {
             self.api
                 .unregister_cluster(name)
                 .await
-                .handle_output(&self.stdout, self.json)
+                .print(&self.stdout, self.json)
         } else {
             Ok(())
         }
@@ -681,6 +683,8 @@ impl StateHub {
             }
         };
 
+        let wait = (locations.len() > 1) || wait;
+
         for location in locations {
             let state = self.api.get_state(&state).await?;
             if state.is_available_in(&location) {
@@ -722,7 +726,7 @@ impl StateHub {
         self.api
             .create_volume(state_name, volume)
             .await
-            .handle_output(&self.stdout, self.json)
+            .print(&self.stdout, self.json)
     }
 
     async fn delete_volume(
@@ -741,7 +745,7 @@ impl StateHub {
                 self.delete_volume_helper(&state, &volume.name, wait)
                     .await
                     .map(Quiet)
-                    .handle_output(&self.stdout, self.json)?;
+                    .print(&self.stdout, self.json)?;
             }
         } else {
             self.verbosely("No such volume")?;
@@ -758,14 +762,14 @@ impl StateHub {
         self.api
             .set_volume_primary(state, volume, primary)
             .await
-            .handle_output(&self.stdout, self.json)
+            .print(&self.stdout, self.json)
     }
 
     async fn list_volumes(&self, state: v0::StateName) -> anyhow::Result<()> {
         self.api
             .get_all_volumes(state)
             .await
-            .handle_output(&self.stdout, self.json)
+            .print(&self.stdout, self.json)
     }
 
     async fn set_availability(self) -> anyhow::Result<()> {
@@ -781,7 +785,7 @@ impl StateHub {
         self.api
             .set_owner(&state, &cluster)
             .await
-            .handle_output(&self.stdout, self.json)
+            .print(&self.stdout, self.json)
     }
 
     async fn unset_owner(
@@ -795,7 +799,7 @@ impl StateHub {
             self.api
                 .unset_owner(&state.name)
                 .await
-                .handle_output(&self.stdout, self.json)
+                .print(&self.stdout, self.json)
         } else {
             anyhow::bail!("Permission denied, you are not theowner of this state.")
         }
@@ -803,7 +807,7 @@ impl StateHub {
 
     pub(crate) fn show<T>(&self, output: Output<T>) -> String
     where
-        T: DeserializeOwned + Serialize + Show,
+        T: de::DeserializeOwned + Serialize + Show,
     {
         output.into_text(self.json)
     }
@@ -844,7 +848,7 @@ impl StateHub {
                     .collect::<HashMap<_, _>>()
             })
             .map(Output::from)
-            .handle_output(&self.stdout, self.json)
+            .print(&self.stdout, self.json)
     }
 
     async fn list_namespaces(&self) -> anyhow::Result<()> {
@@ -914,95 +918,3 @@ impl StateHub {
             .interact_text_on(&self.stdout)
     }
 }
-
-trait HandleOutput {
-    fn handle_output(self, stdout: &Term, json: bool) -> anyhow::Result<()>;
-}
-
-impl<T> HandleOutput for anyhow::Result<T>
-where
-    T: HandleOutput,
-{
-    fn handle_output(self, stdout: &Term, json: bool) -> anyhow::Result<()> {
-        self?.handle_output(stdout, json)
-    }
-}
-
-impl<T> HandleOutput for Output<T>
-where
-    T: DeserializeOwned + Serialize + Show,
-{
-    fn handle_output(self, stdout: &Term, json: bool) -> anyhow::Result<()> {
-        let text = self.into_text(json);
-        stdout.write_line(&text)?;
-        Ok(())
-    }
-}
-
-impl<T> HandleOutput for Detailed<Output<T>>
-where
-    T: DeserializeOwned + Serialize + Show,
-{
-    fn handle_output(self, stdout: &Term, json: bool) -> anyhow::Result<()> {
-        let text = self.into_text(json);
-        stdout.write_line(&text)?;
-        Ok(())
-    }
-}
-
-impl<T> HandleOutput for Quiet<Output<T>>
-where
-    T: DeserializeOwned + Serialize + Show,
-{
-    fn handle_output(self, stdout: &Term, json: bool) -> anyhow::Result<()> {
-        let text = self.into_text(json);
-        stdout.write_line(&text)?;
-        Ok(())
-    }
-}
-
-impl HandleOutput for (Detailed<Output<v0::State>>, Output<Vec<v0::Cluster>>) {
-    fn handle_output(self, stdout: &Term, json: bool) -> anyhow::Result<()> {
-        let (state, clusters) = self;
-        let state_locations = state.all_locations();
-
-        state.handle_output(stdout, json)?;
-
-        let visible_clusters = clusters
-            .iter()
-            .filter(|cluster| {
-                cluster
-                    .all_locations()
-                    .iter()
-                    .any(|location| state_locations.contains(location))
-            })
-            .map(ToString::to_string)
-            .join(" ");
-
-        stdout.write_line(&format!("Visible clusters:\n    {}", visible_clusters))?;
-        Ok(())
-    }
-}
-
-impl HandleOutput for (Detailed<Output<v0::Cluster>>, Output<Vec<v0::State>>) {
-    fn handle_output(self, stdout: &Term, json: bool) -> anyhow::Result<()> {
-        let (cluster, states) = self;
-        let cluster_locations = cluster.all_locations();
-
-        cluster.handle_output(stdout, json)?;
-
-        let visible_states = states
-            .iter()
-            .filter(|state| {
-                state
-                    .all_locations()
-                    .iter()
-                    .any(|location| cluster_locations.contains(location))
-            })
-            .map(ToString::to_string)
-            .join(" ");
-
-        stdout.write_line(&format!("Visible states:\n    {}", visible_states))?;
-        Ok(())
-    }
-}
